/*
   comedi/drivers/act8990.c
   hardware driver for SN74ACT8990 test bus controller based boards 

   COMEDI - Linux Control and Measurement Device Interface
   Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

   EBS - Experimental Boundary Scan project
   Copyright (C) 2002 by CSD, see http://www-csd.ijs.si/applications/BS/BS.html

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
/*
Driver:      act8990.o
Description: device driver for SN74ACT8990 test bus controller based boards
Devices:     (CSD) isa-act8990-ex [isa-act8990-ex]
Author:      Robert Sedevcic <robert.sedevcic@ijs.si>
Status:      testing
Updated:     14.05. 2002
Version:     0.7.0

Configuration options (comedi_config):
  [0] - I/O base port address
        0x200 - 0x22F
        0x240 - 0x26F
        0x2C0 - 0x2EF
        0x300 - 0x32F
        0x340 - 0x36F
        0x3C0 - 0x3EF
  [1] - IRQ number
          3,5,9,10,11 = user defined
          -1          = autodetect  
  [2] - ISA Bus Clock devide 
          1,2,4,8,16

  After module loading board configuration needs to be done. For example:
  $ comedi_config /dev/comedi0 isa-act8990-ex 0x300,-1,1
*/
/*!
  \file act8990.c
  \date 14.05.2002
  \version 0.7.0
  \author Robert Sedevcic <robert.sedevcic@ijs.si>
  \brief This is device driver for SN74ACT8990 test bus controller
  based board (isa-act8990-ex). It's part of EBS (Experimental Boundary Scan)
  project.

  See EBS project home page at http://www-csd.ijs.si/applications/BS/BS.html.

  Copyright (C) 2002 by CSD at http://www-csd.ijs.si
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/comedidev.h>

#include "act8990.h"


/*=============================================================================
 * some device and driver definitions
 *===========================================================================*/
/* array of all supported IRQ numbers */
static int act8990_irq_numbers[] = {-1, -1, -1, 3, -1, 5, -1, -1,
				    -1, 9, 10, 11, -1, -1, -1, -1};

/* array of all supported ISA clock divide numbers */
/*static int act8990_clock_divide[] = {-1, 1, 2, -1, 4, -1, -1, -1,
  8, -1, -1, -1, -1, -1, -1, -1, 16};*/

/* array of SN74ACT8990 based boards */
static act8990_board act8990_boards[] = {
  {
    name: "isa-act8990-ex"
  },
  {
    name: "isa-test"
  }
};

/* driver definition  */
static comedi_driver driver_act8990 = {
  driver_name:  "act8990",
  module:       THIS_MODULE,
  attach:       act8990_attach,
  detach:       act8990_detach,
  /* It is not necessary to implement the following members if you are
   * writing a driver for a ISA PnP or PCI card.
   * Most drivers will support multiple types of boards by
   * having an array of board structures.  These were defined
   * in act8990_boards[] above.  Note that the element 'name'
   * was first in the structure -- Comedi uses this fact to
   * extract the name of the board without knowing any details
   * about the structure except for its length.
   * When a device is attached (by comedi_config), the name
   * of the device is given to Comedi, and Comedi tries to
   * match it by going through the list of board names.  If
   * there is a match, the address of the pointer is put
   * into dev->board_ptr and driver->attach() is called.
   *
   * Note that these are not necessary if you can determine
   * the type of board in software.  ISA PnP, PCI, and PCMCIA
   * devices are such boards.
   */
	board_name:	act8990_boards,
	offset:		sizeof(act8990_board),
	num_names:	sizeof(act8990_boards) / sizeof(act8990_board),
};


/*=============================================================================
 * act8990_interrupt() - interrupt handler function
 *===========================================================================*/
/*!
  \fn static void act8990_interrupt(int irq, void *d, struct pt_regs *regs)
  \brief The act8990_interrupt() is interrupt handler for EBS-act8990-dd.
  \param irq - interrupt request number
  \param d - pointer to comedi_device structure
  \param regs - pointer to pt_regs structure
*/
static void act8990_interrupt(int irq, void *d, struct pt_regs *regs)
{
  sampl_t status0;
  sampl_t status2;
  sampl_t reg_val1;
  sampl_t reg_val2;
  comedi_device *dev;   /* pointer to device */

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_interrupt()\n");
  #endif      

  dev = (comedi_device *)d;

  /* capture Status0/1/2 and Counter1 registers */
  OUT((CAPTURE_STATUS | CAPTURE_COUNTER), MINOR_COMMAND_REG);
  udelay(devprivate->clock_divide);

  /* save the Status0/1 values */
  status0 = IN(STATUS0_REG);
  udelay(devprivate->clock_divide);
  status2 = IN(STATUS2_REG);
  udelay(devprivate->clock_divide);

  #if DEBUG_VAL
    printk("<1>DEBUG_VAL: Status0(1) register in the interrupt is: "
	   "0x%04x\n", status0);
    printk("<1>DEBUG_VAL: Status2(1) register in the interrupt is: "
	   "0x%04x.\n", status2);
    printk("<1>DEBUG_VAL: Counter1 Update0(1) register in the interrupt is:"
	   " 0x%04x.\n", IN(CAPTURE0_REG));
    udelay(devprivate->clock_divide);

    printk("<1>DEBUG_VAL: Counter1 Update1(1) register in the interrupt is:"
	   " 0x%04x.\n", IN(CAPTURE1_REG));
    udelay(devprivate->clock_divide);
  #endif

  /* if both Read registers are full */
  if (status2 & 0x2000) {
    /* get value from the Read register */
    reg_val1 = IN(READ_BUFFER_REG);
    udelay(devprivate->clock_divide);

    #if DEBUG_VAL
      printk("<1>DEBUG_VAL: Both READ registers are full, first value "
             "equals 0x%04x.\n", reg_val1);
    #endif

    /* switch the first and the second bayt */
    reg_val1 = (reg_val1 << 8) | (reg_val1 >> 8);	
    
    if (devprivate->lsb_msb == LSB) {
      /* write first value to read buffer from the back */
      *((sampl_t *)((void *)read_buf->data + devprivate->offset - 
		    read_buf->buf_int_ptr)) = reg_val1;
    } //end LSB
    else { //else MSB
      /* write first value to read buffer from the beginning */
      *((sampl_t *)((void *)read_buf->data + 
		    read_buf->buf_int_ptr)) = reg_val1;
    } //end MSB

    /* two bytes written */
    read_buf->buf_int_count += 2;
    read_buf->buf_int_ptr   += 2;
  } //end both RR full

  /* get value from the Read register */
  reg_val1 = IN(READ_BUFFER_REG);
  udelay(devprivate->clock_divide);

  #if DEBUG_VAL
    printk("<1>DEBUG_VAL: READ register equals 0x%04x.\n", reg_val1);
  #endif

  /* switch the first and the second bayt */
  reg_val1 = (reg_val1 << 8) | (reg_val1 >> 8);

  if (devprivate->lsb_msb == LSB) {
    /* write first value to read buffer from back */
    *((sampl_t *)((void *)read_buf->data + devprivate->offset - 
		  read_buf->buf_int_ptr)) = reg_val1;
  }//end LSB
  else { //else MSB
    /* write first value to read buffer from begining */
    *((sampl_t *)((void *)read_buf->data + 
		  read_buf->buf_int_ptr)) = reg_val1;
  }//end MSB

  /* two bytes written */
  read_buf->buf_int_count += 2;
  read_buf->buf_int_ptr   += 2;

  /* empty(copy to READ register) SHIFTER-FIFO in case of:              */
  /*  -zero passing of COUNTER1 during EXECUTE command                  */
  /*  -zero passing of COUNTER1 during SCAN command                     */
  /*  -the execution of command is cancelled                            */
  /*  -the execution of command is finished                             */
  if (status0 & 0xC030) {
    #if DEBUG_POS
      printk("<1>DEBUG_POS: Counter passed zero, copying fifo to read reg.\n");
    #endif

    /* if FIFO = 32 */
    if (devprivate->shifter_fifo_len == LONG) {
      /* configure Control4 register */
      OUT((devprivate->lsb_msb | 0x0020), CONTROL4_REG);
    }//end FIFO = 32
    else {//else FIFO = 16 
      /* configure Control4 register */
      OUT(devprivate->lsb_msb, CONTROL4_REG);
    }//end FIFO = 16
    udelay(devprivate->clock_divide);

    /* set Counter1 Update0 to 0x000F or 0x001F */
    OUT(devprivate->shifter_fifo_len, COUNTER1_UPDATE0_REG);
    udelay(devprivate->clock_divide);
    /* set Counter1 Update1 register */
    OUT(0x0000, COUNTER1_UPDATE1_REG);
    udelay(devprivate->clock_divide);
    /* update Counter1 register */
    OUT(UPDATE_COUNTER, MINOR_COMMAND_REG);
    udelay(devprivate->clock_divide);
    /* start SCAN minor command */
    OUT(0x5080, MINOR_COMMAND_REG);
    udelay(devprivate->clock_divide);

    /* if FIFO = 32 */
    if (devprivate->shifter_fifo_len == LONG) {
      /* save the Read buffer value */
      reg_val1 = IN(READ_BUFFER_REG);
      udelay(devprivate->clock_divide);

      #if DEBUG_VAL
        printk("<1>DEBUG_VAL: READ register equals 0x%04x.\n", reg_val1);
      #endif

      /* switch the first and the second bayt */
      reg_val1 = (reg_val1 << 8) | (reg_val1 >> 8);
      
      if (devprivate->lsb_msb == LSB) {
	/* write first value to read buffer from back */
	*((sampl_t *)((void *)read_buf->data + devprivate->offset -
		      read_buf->buf_int_ptr)) = reg_val1;
      }//end LSB
      else { //else MSB
	/* write first value to read buffer from begining */
	*((sampl_t *)((void *)read_buf->data + 
		      read_buf->buf_int_ptr)) = reg_val1;
      }//end MSB

      /* two bytes written */
      read_buf->buf_int_count += 2;
      read_buf->buf_int_ptr   += 2;
    }//end FIFO = 32

    /* get value */
    reg_val1 = IN(READ_BUFFER_REG);
    udelay(devprivate->clock_divide);
   
    #if DEBUG_VAL
      printk("<1>DEBUG_VAL: READ register equals 0x%04x.\n", reg_val1);
    #endif

    /* switch the first and the second bayt */
    reg_val1 = (reg_val1 << 8) | (reg_val1 >> 8);

    if (devprivate->lsb_msb == LSB) {
      /* write value to read buffer from back */
      *((sampl_t *)((void *)read_buf->data + devprivate->offset - 
		    read_buf->buf_int_ptr)) = reg_val1;
    }//end LSB
    else { //else MSB
      /* write first value to read buffer from begining */
      *((sampl_t *)((void *)read_buf->data + 
		    read_buf->buf_int_ptr)) = reg_val1;
    }//end MSB

    /* two bytes written */
    read_buf->buf_int_count += 2;
    read_buf->buf_int_ptr   += 2;

    /* restore COUNTER4 register */
    OUT(devprivate->reg[CONTROL4], CONTROL4_REG);
    udelay(devprivate->clock_divide);

    /* reset write buffer pointers */
    write_buf->buf_user_ptr   = 0;
    write_buf->buf_user_count = 0;
    write_buf->buf_int_ptr    = 0;
    write_buf->buf_int_count  = 0;

    /* clear interrupt flags */
    OUT(CLEAR_INT_LOW, MINOR_COMMAND_REG);  
    udelay(devprivate->clock_divide);
    OUT(CLEAR_INT_UP, MINOR_COMMAND_REG);  
    udelay(devprivate->clock_divide);

    /* make read subdevice busy if not yet */
    if (!dev->read_subdev->busy) {
      dev->read_subdev->busy = devprivate->file;
    }

    /* if fifo=32, then throw away 2 first words, else 1 first word */
    if (devprivate->lsb_msb == MSB) {
      read_buf->buf_user_ptr = ((devprivate->shifter_fifo_len + 1)/8) + read_buf->buf_user_ptr; 
    }
    /* if remainder exists */
    if ((devprivate->counter1 + 1) % 16) {
      /* switch the first and the second bayt */
      *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2) = ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2)) << 8) | ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2)) >> 8);
      *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr) = ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr)) << 8) | ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr)) >> 8);
      if (devprivate->lsb_msb == LSB) {
	if (devprivate->counter1 > 15) {
	  if (devprivate->shifter_fifo_len == SHORT) {
	    /* shift and or */
	    *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2) = (*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr) << ((devprivate->counter1 +1) % 16)) | (*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2));
	  }//end FIFO = 16
	  else {//else FIFO = 32
	    *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2) = (*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr) << ((devprivate->counter1+1) % 16)) | ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2)) >> (16 - ((devprivate->counter1 + 1) % 16)));
	  }//end FIFO = 32
	}//end counter1 > 15
	/* set first 16 - %16 bits to 0 */
	*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr) >>= 
	  (16 - ((devprivate->counter1 + 1) % 16));
      }//end LSB
      else { //else MSB
	if (devprivate->counter1 < 0x000F) {//counter1 < 15
	  *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr) = (0xFFFF >> (16-((devprivate->counter1+1) % 16))) & (*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr));
	}//end Counter1 < 15
	else { //else Counter1 > 15
	  *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr) = (0xFFFF >> (16-((devprivate->counter1+1) % 16))) & (*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr));
	}//end counter1 > 15
      }//end MSB
      /* switch the first and the second bayt */
      *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2) = ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2)) << 8) | ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr + 2)) >> 8);
      *(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr) = ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr)) << 8) | ((*(sampl_t *)((void *)read_buf->data + read_buf->buf_user_ptr)) >> 8);
    }//end remainder
    read_buf->buf_user_count += ((devprivate->shifter_fifo_len+1)/8);
  }//end counter passed zero
  /* counter did not passed zero */
  else {
    #ifdef DEBUG_POS    
      printk("<1>Counter1 did not passed zero.\n");
    #endif
    /* get Counter1 value */
    OUT(CAPTURE_COUNTER, MINOR_COMMAND_REG);  
    udelay(devprivate->clock_divide);
    
    reg_val1 = IN(CAPTURE0_REG);
    udelay(devprivate->clock_divide);
    reg_val2 = IN(CAPTURE1_REG);
    udelay(devprivate->clock_divide);

    /* clear interrupt flags */
    OUT(CLEAR_INT_LOW, MINOR_COMMAND_REG);  
    udelay(devprivate->clock_divide);
    OUT(CLEAR_INT_UP, MINOR_COMMAND_REG);  
    udelay(devprivate->clock_divide);

    /* if Counter1 > 15 then write twice to write register, else just once */
    if ((reg_val2 > 0) || (reg_val1 > 15)) {
      /* if LSB format first */
      if (devprivate->lsb_msb == LSB) {
	/* write value from write buffer to write register */
	OUT(*(sampl_t *)((void *)write_buf->data + (write_buf->buf_user_ptr - write_buf->buf_int_ptr - 2)), WRITE_BUFFER_REG);
      }//end LSB
      else {//else MSB
	/* write value from write buffer to write register */
	OUT(*(sampl_t *)((void *)write_buf->data + write_buf->buf_int_ptr),
	    WRITE_BUFFER_REG);
      }//end MSB 
      udelay(devprivate->clock_divide);

      /* two bytes written */
      write_buf->buf_int_ptr   += 2;
      write_buf->buf_int_count += 2;
    }//end two words written

    /* if LSB format first */
    if (devprivate->lsb_msb == LSB) {
      /* write value from write buffer to write register */
      OUT(*(sampl_t *)((void *)write_buf->data + (write_buf->buf_user_ptr - write_buf->buf_int_ptr - 2)), WRITE_BUFFER_REG);
      //  printk("<1>Vrednost vpisana v WR(!!) je 0x%04x\n",*(sampl_t *)((void *)s_write->prealloc_buf - s_write->buf_int_ptr - 2 + s_write->buf_user_ptr));
    }//end LSB
    else {//else MSB
      /* write value from write buffer to write register */
      OUT(*(sampl_t *)((void *)write_buf->data + write_buf->buf_int_ptr), 
	  WRITE_BUFFER_REG);
    }//end MSB 
    udelay(devprivate->clock_divide);

    /* two bytes written */
    write_buf->buf_int_ptr   += 2;
    write_buf->buf_int_count += 2;
  }//end counter != 0 

  /* update Counter1 Update0 and Update1 register */
  #if DEBUG_VAL 
    OUT(UPDATE_COUNTER, MINOR_COMMAND_REG);     
    udelay(devprivate->clock_divide);
    printk("<1>DEBUG_VAL: Counter1 Update0(2) register in the interrupt is:"
	   " 0x%04x.\n", IN(CAPTURE0_REG));
    printk("<1>DEBUG_VAL: Counter1 Update1(2) register in the interrupt is:"
	   " 0x%04x.\n", IN(CAPTURE1_REG));
  #endif

  /* capture STATUS0 register */
  #if DEBUG_VAL
    OUT(CAPTURE_STATUS, MINOR_COMMAND_REG); 
    udelay(devprivate->clock_divide);
    status0 = IN(STATUS0_REG);
    udelay(devprivate->clock_divide);
    printk("<1>DEBUG_VAL: Status0(2) v interrupt is: 0x%04x\n", status0);
  #endif

}//end of act8990_interrupt()


/*=============================================================================
 * act8990_cmd_null() - empty function
 *===========================================================================*/
/*!
  \fn static int act8990_cmd_null(comedi_device *dev, comedi_subdecice *s)
  \brief The act8990_cmd_null() function does nothing
  \param dev - pointer to comedi_device structure
  \param s - pointer to comedi_subdevice structure
  \return int - returns 0
*/
static int act8990_cmd_null(comedi_device *dev, comedi_subdevice *s)
{
  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_cmd_null()\n");
  #endif      

  return 0;
}//end of act8990_cmd_null()


/*=============================================================================
 * act8990_cmd_select() - command select function (pointer on s->do_cmdtest())
 * Switches beetwen MINOR and MAJOR ioctl modes.
 *===========================================================================*/
/*!
  \fn static int act8990_cmd_select(comedi_device *dev, comedi_subdecice *s, comedi_cmd *user_cmd)
  \brief The act8990_cmd_select() switches beetwen minor and major commands
  \param dev - pointer to comedi_device structure
  \param s - pointer to comedi_subdevice structure
  \param user_cmd - pointer to comedi_cmd structure
  \return int - returns 0 if OK or appropriate error otherwise
*/
static int act8990_cmd_select(comedi_device *dev, comedi_subdevice *s,
			      comedi_cmd *user_cmd)
{
  int ret = 0;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_cmd_select()\n");
  #endif

  /* are we busy */
  if (s->busy) {
    dev->write_subdev->busy = NULL;
    dev->read_subdev->busy  = NULL;
    printk("<1>comedi%d: act8990: %s: subdevice busy\n",
	   dev->minor, dev->board_name);
    return -EBUSY;
  }

  /* check for valid file descriptor */
  if (!(current->files->fd[user_cmd->start_arg])) {
    dev->write_subdev->busy = NULL;
    dev->read_subdev->busy  = NULL;
    printk("<1>comedi%d: act8990: %s: not valid file descriptor\n",
	   dev->minor, dev->board_name);
    return -EBADFD;    
  }
  /* save file descriptor */
  s->busy = current->files->fd[user_cmd->start_arg];
  dev->write_subdev->busy = s->busy;
  dev->read_subdev->busy  = s->busy;
  devprivate->file        = s->busy;

  /* pointer to the beginnig of allocated buffer */
  read_buf->data      = read_buf->prealloc_buf;
  read_buf->data_len  = read_buf->prealloc_bufsz;
  write_buf->data     = write_buf->prealloc_buf;
  write_buf->data_len = write_buf->prealloc_bufsz;

  /* save values from user_cmd->data to the devprivate->reg structure */
  act8990_save_regs(dev, user_cmd);
  
  /* update all Control registers */
  act8990_set_control_regs(dev);

  #if DEBUG_VAL
    /* print read/write register values */
  //    act8990_print_regs(dev);
  #endif

  /* switch beetwen two different ioctl modes */
  switch (user_cmd->flags) {
    case MINOR_COM:
      /* Minor Command instructons */
      ret = act8990_minor_command(dev, user_cmd);
      break;
    case MAJOR_COM:
      /* Major Command instructions */
      ret = act8990_major_command(dev, user_cmd);
      break;
    default:
      ret = -EINVAL;
    }

  /* make subdevice unbusy */
  if (s->busy) {
    s->busy = NULL;
  }

  return ret;

}//end of act8990_cmd_select()


/*=============================================================================
 * act8990_attach() is called by the Comedi core to configure the
 * driver for a particular board (device). If you specified a board_name
 * array in the driver structure, dev->board_ptr contains that address.
 * For example:
 *             comedi_config /dev/comedi0 isa-act8990-ex 0x300,-1,1
 * configures the specified board and enters the act8990_attach function if 
 * everything is OK. Configuration options for this board are:
 *   I/O base address  options[0]
 *   IRQ number        options[1]
 *   ISA clock divider options[2]
 *===========================================================================*/
/*!
  \fn static int act8990_attach(comedi_device *dev, comedi_devconfig *it)
  \brief The act8990_attach() allocates structures and other resources
  \param dev - pointer to comedi_device structure
  \param it - pointer to comedi_devconfig structure
  \return int - returns 1 if OK or appropriate error otherwise
*/
static int act8990_attach(comedi_device *dev, comedi_devconfig *it)
{
  short int ret = 0;
  comedi_subdevice *s;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_attach()\n");
  #endif      

  /*
   * If you can probe the device to determine what device in a series
   * it is, this is the place to do it. Otherwise, dev->board_ptr
   * should already be initialized.
   */
  //dev->board_ptr = skel_probe(dev);

  /*
   * Initialize dev->board_name. Use "thisboard" macro.
   */
  dev->board_name = thisboard->name;

  /*
   * Allocate the private structure area. alloc_private()
   * is a convinient macro defined in comedidev.h.
   */
  if ((ret = alloc_private(dev, sizeof(act8990_private))) < 0) {
    return ret;
  }

  /* save ISA clock divider parameter */
  devprivate->clock_divide = 1.5; //act8990_clock_divide[it->options[2]];
  if (devprivate->clock_divide < 0) {
    printk("<1>comedi%d: act8990: %s: not valid ISA clock divider\n",
	   dev->minor, dev->board_name);
    return -EINVAL;
  }

  /* init all hardware parameters */
  if ((ret = act8990_hardware_init(dev, it)) < 0) {
    return ret;
  }

  /*
   * Allocate the subdevice structures.  alloc_subdevice() is a
   * convenient macro defined in comedidev.h.  It relies on
   * n_subdevices being set correctly.
   */
  dev->n_subdevices = SUBDEVICE_NO;

  if ((ret = alloc_subdevices(dev)) < 0) {
    return ret;
  }

  /* the first subdevice is for ioctl */
  s                        = dev->subdevices;
  s->type                  = COMEDI_SUBD_AO;
  s->subdev_flags          = SDF_COMMON; 
  s->n_chan                = 1;
  s->range_table           = &range_bipolar10;
  s->do_cmd                = &act8990_cmd_null;
  s->do_cmdtest            = &act8990_cmd_select;

  /* the second subdevice is writable - from user to kernel space */
  s                        = dev->subdevices + WRITEABLE;
  s->type                  = COMEDI_SUBD_AO;
  s->subdev_flags          = SDF_WRITEABLE; 
  s->n_chan                = 1;
  s->range_table           = &range_bipolar10;
  s->do_cmd                = &act8990_cmd_null;
  s->do_cmdtest            = &act8990_cmd_select;
  dev->write_subdev = s;

  /* the third subdevice is readable - from kernel to user space */
  s                        = dev->subdevices + READABLE;
  s->type                  = COMEDI_SUBD_AI;
  s->subdev_flags          = SDF_READABLE; 
  s->n_chan                = 1;
  s->range_table           = &range_bipolar10;
  s->do_cmd                = &act8990_cmd_null;
  s->do_cmdtest            = &act8990_cmd_select;
  dev->read_subdev = s;

  printk("<1>comedi%d: act8990: version 0.7.0: %s board attached\n",
	 dev->minor, dev->board_name);

  #if DEBUG_DEV
    printk("<1>DEBUG_DEV: Device parameters are:\n");
    printk("<1>  board_name is: %s\n", dev->board_name);
    printk("<1>  iobase port address is: 0x%x\n", dev->iobase);
    printk("<1>  IRQ number is: %d\n", dev->irq);
    printk("<1>  ISA clock divider is: %d\n", devprivate->clock_divide);
  #endif

  return 1;
}//end of act8990_attach()


/*=============================================================================
 * act8990_detach() is called to deconfigure a device. It deallocate
 * resources. This function is also called when act8990_attach fails,
 * so it should be careful not to release resources that were not 
 * necessarily allocated by act8990_attach(). dev->private and 
 * dev->subdevices are deallocated automatically by the core.
 *===========================================================================*/
/*!
  \fn static int act8990_detach(comedi_device *dev)
  \brief The act8990_detach() releases all allocated capabilities before exit.
  \param dev - pointer to comedi_device structure
  \return int - returns 0
*/
static int act8990_detach(comedi_device *dev)
{
  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_detach()\n");
  #endif

  /* release I/O port addresses */
  if (dev->iobase) {
    release_region(dev->iobase, ACT8990_IO_SIZE);
  }

  /* release IRQ number */
  if (dev->irq) {
    free_irq(dev->irq, dev);
  }

  printk("<1>comedi%d: act8990: %s: remove\n",
	 dev->minor, dev->board_name);

  return 0;
}//end of act8990_detach()


/*=============================================================================
 * act8990_hardware_init - init all hardware parameters
 *===========================================================================*/
/*!
  \fn static int act8990_hardware_init(comedi_device *dev, comedi_devconfig *it)
  \brief The act8990_hardware_init() allocates all hardware capabilities that
are needed by the board.
  \param dev - pointer to comedi_device structure
  \param it - pointer to comedi_devconfig structure
  \return int - return 0 if OK or appropriate error otherwise
*/
static int act8990_hardware_init(comedi_device *dev, comedi_devconfig *it)
{
  int iobase;
  short int irq;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_hardware_init()\n");
  #endif

  /* I/O base port address accessing */
  if (it->options[0]) {
    iobase = it->options[0];
  }
  else {
    iobase = ACT8990_IO_DEFAULT;
  }
  if (check_region(iobase, ACT8990_IO_SIZE) < 0) {
    printk("<1>comedi%d: act8990: %s: I/O conflict\n",
	   dev->minor, dev->board_name);
    return -EIO;
  }
  request_region(iobase, ACT8990_IO_SIZE, dev->board_name);
  dev->iobase = iobase;

  /* IRQ accessing */
  if (it->options[1]) {
    irq = it->options[1];
  }
  else {
    irq = -1;
  }
  if (irq < 0) {
    /* autodetecting IRQ number (parameter -1) */
    irq = act8990_autodetect_irq(dev);
  }
  else if (irq < 0x10) {
    /* select one from the list */
    irq = act8990_irq_numbers[irq];
  }
  else {
    /* no irq selected */
    irq = -1;
  }

  if (irq) {
    if (request_irq(irq, act8990_interrupt, SA_INTERRUPT,
		     dev->board_name, dev)) {
      printk("<1>comedi%d: act8990: %s: IRQ not available\n",
	     dev->minor, dev->board_name);
      return -EIO;
    }
  }
  else {//irq < 0
    printk("<1>comedi%d: act8990: %s: IRQ not available\n",
	   dev->minor, dev->board_name);
    return -EIO;
  }
  dev->irq = irq;

  return 0;
}//end of act8990_hardware_init()


/*=============================================================================
 * act8990_autodetect_irq() - autodetecting IRQ number 
 *===========================================================================*/
/*!
  \fn static int act8990_autodetect_irq(comedi_device *dev)
  \brief The act8990_autodetect_irq() autodetects irq number.
  \param dev - pointer to comedi_device structure
  \return int - irq number if detected or -1 otherwise
*/
static int act8990_autodetect_irq(comedi_device *dev)
{
  short int irq = 0;
  unsigned long mask;
  short int count = 0;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_autodetect_irq()\n");
  #endif

  do {
    /* get the unassigned interrupt bitmask */
    mask = probe_irq_on();

    OUT(RESET_ALL, MINOR_COMMAND_REG);       /* reset all ucontroller */
    udelay(devprivate->clock_divide);
    OUT(0x0008, CONTROL0_REG);               /* enable interrupt      */
    udelay(devprivate->clock_divide);
    OUT(0x2008, MINOR_COMMAND_REG);          /* generate interrupt    */
    udelay(devprivate->clock_divide);
    OUT(0x0008, MINOR_COMMAND_REG);          /* disable interrupt     */
    udelay(devprivate->clock_divide);

    /* get the new interrupt mask */
    irq = probe_irq_off(mask);

    if (irq == 0) {
      irq = -1;
    }
  } while ((irq < 0) && ((count++) < 5));

  return irq;
}//enf of act8990_autodetect_irq() 


/*=============================================================================
 * act8990_set_control_regs() - set ACT8990 Control registers
 *===========================================================================*/
/*!
  \fn static void act8990_set_control_regs(comedi_device *dev)
  \brief The act8990_set_control_regs() write values from the devprivate->reg to the control registers.
  \param dev - pointer to comedi_device structure
*/
static void act8990_set_control_regs(comedi_device *dev)
{
  short i, j;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_set_control_regs()\n");    
  #endif

  for (i=0, j=0; i <= 9; i++, j+=0x2) {
    OUT(devprivate->reg[i], j);
    udelay(devprivate->clock_divide);
  }

}//end of act8990_set_control_regs()


/*=============================================================================
 * act8990_save_regs()- save values from user_cmd->data to the devprivate->reg
 *===========================================================================*/
/*!
  \fn static void act8990_save_regs(comedi_device *dev, comedi_cmd *user_cmd)
  \brief The act8990_save_regs() save register values from user_cmd->data to the devprivate->reg.
  \param dev - pointer to comedi_device structure
  \param user_cmd - pointer to comedi_cmd structure
*/
static void act8990_save_regs(comedi_device *dev, comedi_cmd *user_cmd)
{
  short i;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_save_regs()\n");
  #endif

  /* save only read-write registers */
  for (i = 0; i <= 15; i++) {
    devprivate->reg[i] = *(user_cmd->data + i);
  }

}//end of act8990_save_regs()


/*=============================================================================
 * act8990_print_regs() - print the read/write register values
 *===========================================================================*/
/*!
  \fn static void act8990_print_regs(comedi_device *dev)
  \brief The act8990_print_regs() prints read/write register values.
  \param dev - pointer to comedi_device structure
*/
/*static void act8990_print_regs(comedi_device *dev)
{
  short i,j;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_print_regs()\n");
  #endif

  for (i=0, j=0; i <= 21; i++, j+=0x2) {
    printk("<1>DEBUG_VAL: Register %d equals to: 0x%x\n",
	   i , IN(j));
  }

}*/ //end of act8990_print_regs()


/*=============================================================================
 * act8990_minor_command() - minor command ioctl 
 *===========================================================================*/
/*!
  \fn static int act8990_minor_command(comedi_device *dev, comedi_cmd *user_cmd)
  \brief The act8990_minor_command() switches beetwen minor commands
  \param dev - pointer to comedi_device structure
  \param user_cmd - pointer to comedi_cmd structure
  \return int - returns 0
*/
static int act8990_minor_command(comedi_device *dev, comedi_cmd *user_cmd)
{
  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_minor_command()\n");
  #endif

  switch (devprivate->reg[MINOR_COMMAND] & 0xF044) {
    case SET_CONTROL_REGS:
      /* Control registers 0-9 were already loaded,do nothing here */
      #if DEBUG_POS
        printk("<1>DEBUG_POS: SET_CONTROL_REGS activated\n");
      #endif
      break;
    case CAPTURE_STATUS:
    case CAPTURE_COUNTER:
      #if DEBUG_POS
        printk("<1>DEBUG_POS: CAPTURE_STATUS/COUNTER activated\n");
      #endif
      /* set Minor Command register */
      OUT((devprivate->reg[MINOR_COMMAND]
	   | CAPTURE_STATUS | CAPTURE_COUNTER | UPDATE_COUNTER),
	  MINOR_COMMAND_REG);
      udelay(devprivate->clock_divide);
      /* save register values back to the user_cmd->data */
      *(user_cmd->data + STATUS2) = IN(STATUS2_REG); 
      udelay(devprivate->clock_divide);
      *(user_cmd->data + STATUS1) = IN(STATUS1_REG); 
      udelay(devprivate->clock_divide);
      *(user_cmd->data + STATUS0) = IN(STATUS0_REG); 
      udelay(devprivate->clock_divide);
      *(user_cmd->data + CAPTURE0) = IN(CAPTURE0_REG); 
      udelay(devprivate->clock_divide);
      *(user_cmd->data + CAPTURE1) = IN(CAPTURE1_REG); 
      udelay(devprivate->clock_divide);
      break;
    default:
      #if DEBUG_POS
        printk("<1>DEBUG_POS: other Minor Commands activated\n");
      #endif
      /* set Minor Command register */
      OUT(devprivate->reg[MINOR_COMMAND], MINOR_COMMAND_REG);
      udelay(devprivate->clock_divide);
      break;
  }
  return 0;

}//end of act8990_minor_command()


/*=============================================================================
 * act8990_major_command() - major command ioctl 
 *===========================================================================*/
/*!
  \fn static int act8990_major_command(comedi_device *dev, comedi_cmd *user_cmd)
  \brief The act8990_major_command() switches beetwen one of three major commands (move, scan, execute)
  \param dev - pointer to comedi_device structure
  \param user_cmd - pointer to comedi_cmd structure
  \return int - returns 0
*/
static int act8990_major_command(comedi_device *dev, comedi_cmd *user_cmd)
{
  int ret = 0;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: Entering act8990_major_command()\n");
  #endif

  switch (devprivate->reg[MAJOR_COMMAND] & 0xF000) {
    case MOVE_STATE:
      #if DEBUG_POS
        printk("<1>DEBUG_POS: MOVE_STATE major command selected\n");
      #endif
      /* configure Major Command register */	
      OUT(devprivate->reg[MAJOR_COMMAND], MAJOR_COMMAND_REG);
      udelay(devprivate->clock_divide);      
      /* start STATE Major command */
      OUT(0x5080, MINOR_COMMAND_REG);
      udelay(devprivate->clock_divide);            
      break;
    case START_SCAN:
      ret = act8990_start_scan(dev);
      break;
    case START_EXECUTE:
      ret = act8990_start_execute(dev);
      break;
  }

  return ret;

}//end of act8990_major_command()


/*=============================================================================
 * act8990_start_execute - start execute major command
 *===========================================================================*/
/*!
  \fn static int act8990_start_execute(comedi_device *dev)
  \brief The act8990_start_execute() starts execute major command.
  \param dev - pointer to comedi_device structure
  \return int - always returns 0
*/
static int act8990_start_execute(comedi_device *dev)
{
  #if DEBUG_POS
    printk("<1>DEBUG_POS: act8990_start_execute() activated\n");
  #endif

  /* save Counter1 (Update0 and Update1 register) */
  devprivate->counter1 = (((devprivate->reg[COUNTER1_UPDATE1]) << 16) |
	      (devprivate->reg[COUNTER1_UPDATE0]));

  /* load Counter1 Update0 register */
  OUT(devprivate->reg[COUNTER1_UPDATE0], COUNTER1_UPDATE0_REG);
  udelay(devprivate->clock_divide);      
  /* load Counter1 Update1 register */
  OUT(devprivate->reg[COUNTER1_UPDATE1], COUNTER1_UPDATE1_REG);
  udelay(devprivate->clock_divide);      

  /* update Counter1 registers */
  OUT(UPDATE_COUNTER, MINOR_COMMAND_REG);
  udelay(devprivate->clock_divide);      

  /* select execute major command */
  OUT(devprivate->reg[MAJOR_COMMAND], MAJOR_COMMAND_REG);
  udelay(devprivate->clock_divide);      

  /* start execute minor command */
  OUT(0x408F, MINOR_COMMAND_REG);
  udelay(devprivate->clock_divide);      

  return 0;
}//end of act8990_start_execute()


/*=============================================================================
 * act8990_start_scan - start scan major command
 *===========================================================================*/
/*!
  \fn static int act8990_start_scan(comedi_device *dev)
  \brief The act8990_start_scan() starts scan major command.
  \param dev - pointer to comedi_device structure
  \return int - always returns 0
*/
static int act8990_start_scan(comedi_device *dev)
{
  sampl_t status;
  unsigned int counter1;

  #if DEBUG_POS
    printk("<1>DEBUG_POS: act8990_start_scan() activated\n");
  #endif

  /* extract the Counter1 value from Update0 and Update1 - in bits */
  counter1 = (((devprivate->reg[COUNTER1_UPDATE1]) << 16) |
	      (devprivate->reg[COUNTER1_UPDATE0]));

  devprivate->counter1 = counter1;

  /* set bytes number */
  devprivate->nbytes = ((counter1 + 1) / 16) * 2;
  if ((counter1 + 1) % 16) {
    devprivate->nbytes += 2;
  }


  /* set write buffer interrupt offsets to 0 */
  write_buf->buf_int_ptr   = 0;
  write_buf->buf_int_count = 0;

  /* write buffer should not be empty */
  if (write_buf->buf_int_ptr >= write_buf->buf_user_ptr) {
    printk("<1>comedi%d: act8990: %s: no data in write buffer\n",
	   dev->minor, dev->board_name);
    return -ENODATA;
  }

  /* Counter1 should not be bigger than number of written data (in bits) */
  if ((counter1 + 1) > ((write_buf->buf_user_ptr - write_buf->buf_int_ptr)*8)) {
    printk("<1>comedi%d: act8990: %s: counter1 bigger than written data\n",
	   dev->minor, dev->board_name);
    return -EINVAL;
  }

  /* set LSB or MSB transfer format */
  if ((devprivate->reg[CONTROL4] & 0x0800)) {
    devprivate->lsb_msb = MSB;
  }
  else {
    devprivate->lsb_msb = LSB;
  }

  /* set Shifter-fifo length */
  if (devprivate->reg[CONTROL4] & 0x0020) {
    devprivate->shifter_fifo_len = LONG;         /* 32 bits long */
    devprivate->offset = devprivate->nbytes + 2;
  }
  else {
    devprivate->shifter_fifo_len = SHORT;        /* 16 bits long */
    devprivate->offset = devprivate->nbytes;    
  }

  /* load Counter1 Update0 register */
  OUT(devprivate->reg[COUNTER1_UPDATE0], COUNTER1_UPDATE0_REG);
  udelay(devprivate->clock_divide);      
  /* load Counter1 Update1 register */
  OUT(devprivate->reg[COUNTER1_UPDATE1], COUNTER1_UPDATE1_REG);
  udelay(devprivate->clock_divide);      

  /* update Counter1 registers */
  OUT(UPDATE_COUNTER, MINOR_COMMAND_REG);
  udelay(devprivate->clock_divide);      

  /* select scan major command */
  OUT(devprivate->reg[MAJOR_COMMAND], MAJOR_COMMAND_REG);
  udelay(devprivate->clock_divide);      

  /* start scan minor command */
  OUT(0x5080, MINOR_COMMAND_REG);
  udelay(devprivate->clock_divide);

  /* set all read buffer offsets to 0 */
  read_buf->buf_int_ptr    = 0;
  read_buf->buf_int_count  = 0;
  read_buf->buf_user_ptr   = 0;
  read_buf->buf_user_count = 0;

  /* refresh Status and Counter1 registers */
  OUT((CAPTURE_STATUS | CAPTURE_COUNTER), MINOR_COMMAND_REG);
  udelay(devprivate->clock_divide);

  /* save Status2 register value */
  status = IN(STATUS2_REG); 
  udelay(devprivate->clock_divide);

  /* reset data registers if not empty */
  if ((status & 0x1000) || (status & 0x0200)) {
    OUT(RESET_DATA, MINOR_COMMAND_REG);
    udelay(devprivate->clock_divide);
  }

  /* if Counter1 > 15 then write to write buffer twice, else once */
  if (counter1 > 15) {
    /* if LSB format first */
    if (devprivate->lsb_msb == LSB) {
      OUT(*(sampl_t *)((void *)write_buf->data + (write_buf->buf_user_ptr - write_buf->buf_int_ptr - 2)),
	  WRITE_BUFFER_REG);
    }//end of LSB
    /* MSB format first */
    else {
      OUT(*(sampl_t *)((void *)write_buf->data + write_buf->buf_int_ptr),
	  WRITE_BUFFER_REG);
    }//end of MSB
    udelay(devprivate->clock_divide);

    /* increment by 2, 2 bytes written */
    write_buf->buf_int_ptr += 2;
    write_buf->buf_int_count += 2;

  }//end of Counter1 > 15

  /* if LSB format first */
  if (devprivate->lsb_msb == LSB) {
    OUT(*(sampl_t *)((void *)write_buf->data + (write_buf->buf_user_ptr - write_buf->buf_int_ptr - 2)),
	WRITE_BUFFER_REG);
  }//end of LSB
  /* MSB format first */
  else {
    OUT(*(sampl_t *)((void *)write_buf->data + write_buf->buf_int_ptr),
	WRITE_BUFFER_REG);
  }//end of MSB
  udelay(devprivate->clock_divide);

  /* increment by 2, 2 bytes written */
  write_buf->buf_int_ptr += 2;
  write_buf->buf_int_count += 2;

  /* if both pointers are equal, then reset them */
  if (write_buf->buf_user_ptr == write_buf->buf_int_ptr) {
    write_buf->buf_user_ptr = 0;
    write_buf->buf_int_ptr  = 0;
  }

  return 0;

}//end of act8990_start_scan()


COMEDI_INITCLEANUP(driver_act8990);
