/*
   comedi/drivers/act8990.h
   hardware driver for SN74ACT8990 test bus controller based boards 

   COMEDI - Linux Control and Measurement Device Interface
   Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

   EBS - Experimental Boundary Scan project
   Copyright (C) 2002 by CSD, see http://www.csd-ijs.si/projects/BS/

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
/*
Driver:      act8990.o
Description: device driver for SN74ACT8990 test bus controller based boards
Devices:     (CSD) isa-act8990-ex [isa-act8990-ex]
Author:      Robert Sedevcic <robert.sedevcic@ijs.si>
Status:      in progres
Updated:     05.04. 2002
Version:     0.7.0

Configuration options (comedi_config):
  [0] - I/O base port address
        0x200 - 0x22F
        0x240 - 0x26F
        0x2C0 - 0x2EF
        0x300 - 0x32F
        0x340 - 0x36F
        0x3C0 - 0x3EF
  [1] - IRQ number
          3,5,9,10,11 = user defined
          -1          = autodetect  
  [2] - ISA Bus Clock devide 
          1,2,4,8,16

  After module loading board configuration needs to be done. For example:
  $ comedi_config /dev/comedi0 isa-act8990-ex 0x300,-1,1
*/
/*!
  \file act8990.h
  \date 14.05.2002
  \version 0.7.0
  \author Robert Sedevcic <robert.sedevcic@ijs.si>
  \brief This is device driver for SN74ACT8990 test bus controller
  based board (isa-act8990-ex). It's part of EBS (Experimental Boundary Scan)
  project.

  See EBS project home page at http://www-csd.ijs.si/applications/BS/BS.html.

  Copyright (C) 2002 by CSD at http://www-csd.ijs.si
 */

#ifndef _ACT8990_H_
#define _ACT8990_H_


#ifndef ACT8990_REGS_NO
  #define ACT8990_REGS_NO 24 /* ACT8990 register number */
#endif


/*=============================================================================
 * structure declarations
 *===========================================================================*/
/*! \struct act8990_private
    \brief act8990 private data structure
*/
typedef struct act8990_private_struct {
  void *file;                     //!< pointer to temporary open file str.
  short clock_divide;             //!< ISA clock divider                  
  sampl_t reg[ACT8990_REGS_NO];   //!< temporary registers storage       
  sampl_t lsb_msb;                //!< lsb or msb send and received first 
  unsigned int counter1;          //!< Counter1 value                     
  unsigned int nbytes;            //!< bytes number                       
  unsigned int offset;            //!< offset number                      
  unsigned int shifter_fifo_len;  //!< Shifter-fifo length (16/32)        
}act8990_private;

/*! \struct act8990_board
    \brief SN74ACT8990 based board structure
*/
typedef struct act8990_board_struct {
  char *name;     //!< board name
}act8990_board;


/*=============================================================================
 * board - device definitions
 *===========================================================================*/
#ifndef ACT8990_IO_DEFAULT
  #define ACT8990_IO_DEFAULT 0x300  /* default I/O base address */
#endif

#ifndef ACT8990_IO_SIZE
  #define ACT8990_IO_SIZE 48        /* I/O port size 2*24 = 48 */
#endif

/* shorthand access to device private structure */
#ifndef devprivate
  #define devprivate ((act8990_private *)(dev->private))
#endif

/* shorthand access to the particular board structure,
   each board gets its own dev structure */
#ifndef thisboard
  #define thisboard ((act8990_board *)dev->board_ptr)
#endif

/* shorthand access to the write buffer parameters */
#ifndef write_buf
  #define write_buf dev->write_subdev->async
#endif

/* shorthand access to the read buffer parameters */
#ifndef read_buf
  #define read_buf dev->read_subdev->async
#endif

/* subdevice number */
#ifndef SUBDEVICE_NO
  #define SUBDEVICE_NO 3
#endif

/* subdevice is writable - from user to kernel space */
#ifndef WRITEABLE
  #define WRITEABLE 1
#endif

/* subdevice is readable - from kernel to user space */
#ifndef READABLE
  #define READABLE 2
#endif


/* msb bit first in and out */
#ifndef MSB
  #define MSB 0x0E00   
#endif

/* lsb bit first in and out */
#ifndef LSB
  #define LSB 0x0600   
#endif

/* shifter-fifo is 31+1 bits long */
#ifndef LONG
  #define LONG 0x001F   
#endif

/* shifter-fifo is 15+1 bits long */
#ifndef SHORT
  #define SHORT 0x000F   
#endif


/*=============================================================================
 * driver definitions
 *===========================================================================*/
#ifndef OUT
  #define OUT(value, register) outw_p(value, dev->iobase + register)
#endif

#ifndef IN
  #define IN(register) inw_p(dev->iobase + register)
#endif

#ifndef DEBUG_DEV
  #define DEBUG_DEV 0
#endif

#ifndef DEBUG_POS
  #define DEBUG_POS 0
#endif

#ifndef DEBUG_VAL
  #define DEBUG_VAL 0
#endif


/*=============================================================================
 * some Minor Command predefined register values
 *===========================================================================*/
#ifndef RESET_ALL
  #define RESET_ALL 0x7001   /* reset the whole controller */
#endif

#ifndef RESET_IRQF
  #define RESET_IRQF 0x7002  /* reset the interrupt request flags */
#endif

#ifndef RESET_DATA
  #define RESET_DATA 0x7004  /* reset the data registers */
#endif

#ifndef RESET_REQF
  #define RESET_REQF 0x7008  /* reset the request flags */
#endif

#ifndef CAPTURE_STATUS
  #define CAPTURE_STATUS 0x6040  /* capture Status0,1,2 register values */
#endif

#ifndef CAPTURE_COUNTER
  #define CAPTURE_COUNTER 0x6004 /* capture Counter1 register */
#endif

#ifndef UPDATE_COUNTER
  #define UPDATE_COUNTER 0x6001 /* update Counter1 register */
#endif

#ifndef SET_CONTROL_REGS
  #define SET_CONTROL_REGS 0x8000
#endif

#ifndef CLEAR_INT_LOW
  #define CLEAR_INT_LOW 0x003F  /* clear lower interrupt */
#endif

#ifndef CLEAR_INT_UP
  #define CLEAR_INT_UP 0x10FF   /* clear upper interrupt */
#endif


/*=============================================================================
 * user_cmd.flags allowed parameters
 *===========================================================================*/
#ifndef MINOR_COM
  #define MINOR_COM 0        /* Minor Command instructions selection */
#endif

#ifndef MAJOR_COM
  #define MAJOR_COM 1        /* Major Command instructions selection */
#endif


/*=============================================================================
 * some Major Command predefined register values
 *===========================================================================*/
#ifndef MOVE_STATE
  #define MOVE_STATE 0x1000
#endif

#ifndef TEST_LOGIC_RESET
  #define TEST_LOGIC_RESET 0x1001
#endif

#ifndef RUN_TEST_IDLE
  #define RUN_TEST_IDLE 0x1002
#endif

#ifndef SHIFT_DR
  #define SHIFT_DR 0x1004
#endif

#ifndef SHIFT_IR
  #define SHIFT_IR 0x1005
#endif

#ifndef PAUSE_DR
  #define PAUSE_DR 0x1006
#endif

#ifndef PAUSE_IR
  #define PAUSE_IR 0x1007
#endif


#ifndef START_EXECUTE
  #define START_EXECUTE 0x20CF
#endif


#ifndef START_SCAN
  #define START_SCAN 0x3000
#endif


/*=============================================================================
 * function prototypes
 *===========================================================================*/
/* The comedi_driver structure tells the Comedi core module
   which function to call to configure/deconfigure (attach/detach)
   the board and also about the kernel module that contains
   the device code
*/
static int act8990_attach(comedi_device *dev, comedi_devconfig *it);
static int act8990_detach(comedi_device *dev);

static int act8990_autodetect_irq(comedi_device *dev);
static int act8990_hardware_init(comedi_device *dev, comedi_devconfig *it);

static void act8990_save_regs(comedi_device *dev, comedi_cmd *user_cmd);
//static void act8990_print_regs(comedi_device *dev);
static void act8990_set_control_regs(comedi_device *dev);

static int act8990_minor_command(comedi_device *dev, comedi_cmd *user_cmd);
static int act8990_major_command(comedi_device *dev, comedi_cmd *user_cmd);

static int act8990_start_execute(comedi_device *dev);
static int act8990_start_scan(comedi_device *dev);


/*=============================================================================
 * register array offsets
 *===========================================================================*/
#define CONTROL0   0
#define CONTROL1   1
#define CONTROL2   2
#define CONTROL3   3
#define CONTROL4   4
#define CONTROL5   5
#define CONTROL6   6
#define CONTROL7   7
#define CONTROL8   8
#define CONTROL9   9

#define MINOR_COMMAND 10
#define MAJOR_COMMAND 11

#define COUNTER1_UPDATE0 12
#define COUNTER1_UPDATE1 13
#define COUNTER2_UPDATE0 14
#define COUNTER2_UPDATE1 15
#define CAPTURE0 16
#define CAPTURE1 17

#define STATUS0 18
#define STATUS1 19
#define STATUS2 20
#define STATUS3 21

#define READ_BUFFER  22
#define WRITE_BUFFER 23


/*=============================================================================
 * ACT8990 register offset addresses 
 *===========================================================================*/
/*=========================================================================== 
 *    Control Registers
 *      Reg. name      Offset    Application                        Access
 *===========================================================================*/
#define CONTROL0_REG   0x00      /* Interrupt enable control        (R/W)*/
#define CONTROL1_REG   0x02      /* Host and test control           (R/W)*/
#define CONTROL2_REG   0x04      /* TMS0-TMS5 connections           (R/W)*/
#define CONTROL3_REG   0x06      /* TMS/EVENT, TDO, TMS/TCK control (R/W)*/
#define CONTROL4_REG   0x08      /* Delay, data format of TDI, TDO  (R/W)*/
#define CONTROL5_REG   0x0A      /* Event pins data out., direction (R/W)*/
#define CONTROL6_REG   0x0C      /* Event choice of Event0/1 inputs (R/W)*/
#define CONTROL7_REG   0x0E      /* Event choice of Event2/3 inputs (R/W)*/
#define CONTROL8_REG   0x10      /* Counter1 control                (R/W)*/
#define CONTROL9_REG   0x12      /* Counter2 control                (R/W)*/   

/*===========================================================================
 *    Command Registers
 *      Reg. name         Offset Application                        Access
 *===========================================================================*/
#define MINOR_COMMAND_REG 0x14   /* Interrupt, read control, ...    (R/W)*/
#define MAJOR_COMMAND_REG 0x16   /* State, Execute, Scan control    (R/W)*/

/*===========================================================================
 *    Counters control Registers
 *      Reg. name            Offset Application                     Access
 *===========================================================================*/
#define COUNTER1_UPDATE0_REG 0x18  /* Counter1 lower value word     (R/W)*/
#define COUNTER1_UPDATE1_REG 0x1A  /* Counter1 higher value word    (R/W)*/
#define COUNTER2_UPDATE0_REG 0x1C  /* Counter2 lower word of Cou.20 (R/W)*/
#define COUNTER2_UPDATE1_REG 0x1E  /* Counter2 lower word of Cou.21 (R/W)*/
#define CAPTURE0_REG 0x28          /* Counter1 lower value word     (R)  */
#define CAPTURE1_REG 0x2A          /* Counter1 higher value word    (R)  */ 

/*============================================================================
 *    Status Registers
 *      Reg. name   Offset         Application                      Access
 *===========================================================================*/
#define STATUS0_REG 0x20           /* Interrupt flag status         (R)  */
#define STATUS1_REG 0x22           /* External signals and request  (R)  */
#define STATUS2_REG 0x24           /* Board and buffer state        (R)  */  
#define STATUS3_REG 0x26           /* Not used                      (R)  */

/*============================================================================
 *    Read/Write Registers
 *      Reg. name         Offset   Application                      Access
 *==========================================================================*/ 
#define READ_BUFFER_REG  0x2C      /* Input buffer for TDI          (R)  */
#define WRITE_BUFFER_REG 0x2E      /* Output buffer for TDO         (W)  */
/*===========================================================================*/

#endif
